/*----------------------------------------------------------------------------
EXPORTS:
int cg_sprintf(char **dst_ptr, const char *fmt, ...);
int str_cat(char **left_ptr, const char *right);

Highly stripped-down printf(). No floating point, no field widths,
no modifier characters. Only %%, %d, %i, %u, %c, %s are supported.
Only cg_sprintf() is implemented; a function that works like asprintf()
but it _appends_ it's output to the supplied buffer.
----------------------------------------------------------------------------*/
#include <string.h> /* strlen(), strncpy() */
#include <stdarg.h> /* va_list, va_start(), va_arg(), va_end() */
#include <stdio.h> /* EOF */
/*****************************************************************************
Converts unsigned binary numeric value 'num' to ASCII string.
'buf' must point to the END of a suitably-large buffer.
Return value is a pointer to the START of the string stored in 'buf'.
*****************************************************************************/
static char *num2asc(char *buf, unsigned radix, unsigned long num)
{
	unsigned j;

/* do..while() instead of while() ensures we get a single '0' if num==0 */
	do
	{
		j = (unsigned)(num % radix);
		buf--;
		if(j < 10)
			*buf = j + '0';
		else
			*buf = j - 10 + 'A';
		num /= radix;
	} while(num != 0);
	return buf;
}
/*****************************************************************************
name:	do_printf
action:	minimal subfunction for ?printf, calls function
	'fn' with arg 'ptr' for each character to be output
returns:total number of characters output
*****************************************************************************/
typedef int (*pfnptr_t)(unsigned c, void **helper);

static int do_printf(const char *fmt, va_list args, pfnptr_t fn, void *ptr)
{
/* Largest number handled is 2^32-1, lowest radix handled is 10.
2^32-1 in base 10 has 9 digits. Add 7 for trailing zero byte and for slop.
Strings are copied directly to the output, and don't use 'buf'. */
#define	PR_BUFLEN	16
	char negative, *where, buf[PR_BUFLEN];
	unsigned count;
	long num;
	int i;

	for(count = 0; *fmt != '\0'; fmt++)
	{
		if(*fmt != '%')
		{
			if((i = fn(*fmt, &ptr)) == EOF)
				return i;
			count++;
			continue;
		}
		fmt++;
/* %% */
		if(*fmt == '%')
		{
			if((i = fn(*fmt, &ptr)) == EOF)
				return i;
			count++;
			continue;
		}
		negative = 0;
		where = &buf[PR_BUFLEN - 1];
		*where = '\0';
/* numeric conversions */
		if(*fmt == 'd' || *fmt == 'i' || *fmt == 'u')
		{
			if(*fmt == 'd' || *fmt == 'i')
			{
				num = va_arg(args, signed int);
/* convert to unsigned but remember that it was negative */
				if(num < 0)
				{
					negative = 1;
					num = -num;
				}
			}
			else if(*fmt == 'u')
				num = va_arg(args, unsigned int);
/* convert binary unsigned number to decimal ASCII */
			where = num2asc(where, 10, num);
		}
/* convert char to string for output */
		else if(*fmt == 'c')
		{
			where--;
			*where = (unsigned char)va_arg(args, unsigned char);
		}
		else if(*fmt == 's')
			where = va_arg(args, char *);
/* bogus conversion character -- copy it to output and go back to state 0 */
		else
		{
			if((i = fn(*fmt, &ptr)) == EOF)
				return i;
			count++;
			continue;
		}
/* emit converted number/char/string */
		if(negative)
		{
			if((i = fn('-', &ptr)) == EOF)
				return i;
			count++;
		}
		for(; *where != '\0'; where++)
		{
			if((i = fn(*where, &ptr)) == EOF)
				return i;
			count++;
		}
	}
	return count;
}
/****************************************************************************/
#include <stdlib.h> /* realloc() */

typedef struct
{
	char *buf;
	unsigned alloc, ptr;
} asprintf_t;

static int vasprintf_help(unsigned c, void **ptr)
{
	asprintf_t *as;
	unsigned len;
	char *t;

	as = *ptr;
	if(as->ptr + 1 >= as->alloc)
	{
/* start with 16 bytes,
then double the string length every time it overflows */
		len = (as->alloc == 0) ? 16 : (as->alloc * 2);
		if((t = realloc(as->buf, len)) == NULL)
			return EOF; // xxx - return some other value?
		as->buf = t;
		as->alloc = len;
	}
	as->buf[as->ptr] = c;
	as->ptr++;
	return 0 ;
}
/****************************************************************************/
static int vcg_sprintf(char **dst_ptr, const char *fmt, va_list args)
{
	asprintf_t as;
	int i;

	if(dst_ptr == NULL)
		return EOF;	// xxx - return some other value?
	as.buf = *dst_ptr;
	as.ptr = as.alloc = (as.buf == NULL) ? 0 : strlen(as.buf);
	if((i = do_printf(fmt, args, vasprintf_help, &as)) != EOF)
	{
		as.buf[as.ptr] = '\0';
		*dst_ptr = as.buf;
	}
	return i;
}
/******************************************************************************
No buffer overflow here, hopefully -- a buffer of sufficient size
should be allocated by vasprintf_help(). The value returned via dst_ptr
should be free()d when no longer needed.
******************************************************************************/
int cg_sprintf(char **dst_ptr, const char *fmt, ...)
{
	va_list args;
	int rv;

	va_start(args, fmt);
	rv = vcg_sprintf(dst_ptr, fmt, args);
	va_end(args);
	return rv;
}
/*****************************************************************************
Appends (conCATenates) C string 'right' to variable-length string 'left_ptr'
Returns 0 if success, <0 if error
*****************************************************************************/
int str_cat(char **left_ptr, const char *right)
{
	unsigned left_len, right_len, tot_len;
	char *left;

	if(left_ptr == NULL)
		return -1;	// xxx - return some other value?
	left_len = (*left_ptr != NULL) ? strlen(*left_ptr) : 0;
	right_len = strlen(right) + 1;
	tot_len = left_len + right_len;
	if((left = realloc(*left_ptr, tot_len)) == NULL)
		return -1;	 // xxx - return some other value?
	strncpy(&left[left_len], right, right_len);
	*left_ptr = left;
	return tot_len;
}
